#include "OpeTool.h"
#include "SctData/RawScanResult.h"
#include "Sct/LogicErrors.h"
#include <boost/shared_ptr.hpp>

#include "TH2.h"
#include "TH1.h"
#include "TMath.h"
#include "TFile.h"

using namespace SctData;
using namespace Sct;
using boost::shared_ptr;

namespace SctAnalysis {

  OpeTool::OpeTool(){}

  OpeTool::~OpeTool(){}

  boost::shared_ptr<SctData::OpeResult> OpeTool::analyzeModule(const RawScanResult& raw){

    unsigned nOPE = raw.nOccupancyPerEvent();

    shared_ptr<OpeResult> result;
    if (nOPE==0) return result;
    
    result = shared_ptr<OpeResult>(new OpeResult(raw.getUniqueID()));

    //For each chip... 
    for (unsigned int ichip=0; ichip<nOPE; ++ichip) {
      const TH2D& chipHist = raw.getOccupancyPerEvent(ichip);
      const int biny = chipHist.GetNbinsY();

      std::ostringstream newhistname;
      newhistname << raw.getClassName() << "." << raw.getUniqueID() 
		  << "_Chip" << ichip << "_Ope_Badness"; 

      std::ostringstream newtitle;
      newtitle << "Chip" << ichip << "_Ope_Badness"; 

      const char* c_name=newhistname.str().c_str();
      const char* c_title=newtitle.str().c_str();
      // due to Brun's incompetence we need to make a non-const number of bins :(
      int stupidBloomingNonConstBinY=biny;
      shared_ptr<TH1> badnessHist(new TH1D(c_name, c_title, stupidBloomingNonConstBinY, chipHist.GetYaxis()->GetXbins()->GetArray()));

      // loop over y bins making projections
      for(unsigned int iy=1; iy<biny+1; ++iy){
	std::ostringstream projectionName;
	projectionName << raw.getClassName() << "." << raw.getUniqueID() 
		       << "_Chip" << ichip << "_bin" << iy; 
	shared_ptr<TH1D> proj(chipHist.ProjectionX(projectionName.str().c_str(),iy,iy,""));
	//	std::cout << "OPE: Bin " << iy << " ...." 
	//	  << findChi2Binomial(*proj) << " fraction :"  << fractionVariance(*proj) <<std::endl;
	// std::cout << "OPE: Chip: " << ichip << " Bin " << iy << "threshold: " << chipHist.GetYaxis()->GetBinCenter(iy) <<" fraction :"  << fractionVariance(*proj) <<std::endl;
	badnessHist->SetBinContent(iy, fractionVariance(*proj));	
      }

      SctData::ChipOpeResult& chipResult=result->getChipResult(ichip);

      // FILL IN THAT CHIP + give max
      chipResult.hist=badnessHist;
      chipResult.badmax= badnessHist->GetMaximum();
      int badbin = badnessHist->GetMaximumBin();
      chipResult.threshold= badnessHist->GetBinCenter(badbin);
    } 
    
    return result;
  }

  double OpeTool::fractionVariance(const TH1& data) {

    double range = data.GetXaxis()->GetXmax() - data.GetXaxis()->GetXmin();
    double prob = (data.GetMean())/range;
    // theoretically calculate the RMS from the probability dist:
    double calcRMS = sqrt(range*prob*(1-prob));
    // measure the RMS from data
    double measRMS = data.GetRMS();
    // return the fraction of the measured and calculated RMS
    // std::cout << "OPE: measuredRMS :" << measRMS << " calcRMS :" << calcRMS << std::endl;	
    if (calcRMS ==0) return 1;
    else return measRMS/calcRMS;
  }
  
  double OpeTool::findChi2Binomial(const TH1& data) {

    const unsigned int binx = data.GetNbinsX();
    unsigned int dof = 0;
    double trials = data.GetSumOfWeights();
    double range = data.GetXaxis()->GetXmax() - data.GetXaxis()->GetXmin();
    if (range==0.) throw Sct::LogicError("Range of histogram x axis was zero!",__FILE__,__LINE__);
    double prob = (data.GetMean())/range;
    std::cout << "OPE: Mean:" << data.GetMean()<< std::endl;	

    double chisq =0;
    double delta =0;

    if (prob == 0) return 0;
    else {
      // Root counts bin numbers starting from 1
      // the first bin goes from -0.5 to 0.5 -- no physical meaning 
      // because a trigger requires an event... 
      // the following bins (0.5,4.5), (4.5,8.5) and so on... 
      for (unsigned int ix=1; ix<binx+1; ix++){
	
	// to make the computation easier and faster, calculate only for those bins which have data in it... 
	if (data.GetBinContent(ix) > 0) {
	  
	  // want to sum over all ints greater than (the first int above the lower bin edge) 
	  // but less than (the first int above the upper bin edge)
	  int start = static_cast<int>(ceil(data.GetBinLowEdge(ix)));
	  int w = int(data.GetBinWidth(ix));
	  if (w ==0) throw Sct::LogicError("Error- bin size truncates to zero!", __FILE__, __LINE__);
	  std::cout << "Giving BinomialI " << prob << "range: " << start << " to " << start+w << std::endl;
	  int irange=static_cast<int>(floor(range+0.5));
	  double expect =  (BinomialI(prob,irange,start) - 
			    BinomialI(prob,irange,start+w))*trials;
	  
	  if (expect ==0) delta = data.GetBinContent(ix);
	  else delta  = (data.GetBinContent(ix)-expect)/sqrt(expect);
	  
	  chisq += delta*delta;
	  std::cout << "OPE: expect:" << expect<< " get: "<<data.GetBinContent(ix) << " delta:" << delta << std::endl;	
	  dof += 1;
	}
      }
      // binomial distribution has one empty parameter.. 
      dof = dof -1;
      if (dof == 0 ) return 0;
      else return sqrt(chisq)/(dof-1);
    }
  }
  // The following has been copied from the latest ROOT libraries since the 
  // installed ROOT libraries are not necessarily the latest
  double OpeTool::BinomialI(Double_t p, Int_t n, Int_t k)
    {
      // Suppose an event occurs with probability _p_ per trial
      // Then the probability P of its occuring _k_ or more times
      // in _n_ trials is termed a cumulative binomial probability
      // the formula is P = sum_from_j=k_to_n(TMath::Binomial(n, j)*
      // *TMath::Power(p, j)*TMath::Power(1-p, n-j)
      // For _n_ larger than 12 BetaIncomplete is a much better way
      // to evaluate the sum than would be the straightforward sum calculation
      // for _n_ smaller than 12 either method is acceptable
      // ("Numerical Recipes")
      //     --implementation by Anna Kreshuk
      
      Double_t P = BetaIncomplete(p, Double_t(k), Double_t(n-k+1));
      return P;
    }
  
  Double_t OpeTool::BetaIncomplete(Double_t x, Double_t a, Double_t b)
    {
      // Calculates the incomplete Beta-function.
      //  -- implementation by Anna Kreshuk
      
      Double_t bt;
      if ((x<0.0)||(x>1.0)) {
	throw Sct::LogicError("OpeTool::BetaIncomplete: X must between 0 and 1", __FILE__, __LINE__);
	return 0.0;
      }
      if ((x==0.0)||(x==1.0)) {
	bt=0.0;
      } else {
	bt = TMath::Power(x, a)*TMath::Power(1-x, b)/Beta(a, b);
      }
      if (x<(a+1)/(a+b+2)) {
	return bt*BetaCf(x, a, b)/a;
      }
      else {
	return (1 - bt*BetaCf(1-x, b, a)/b);
      }
    }
  
 Double_t OpeTool::Beta(Double_t p, Double_t q)
{
   // Calculates Beta-function Gamma(p)*Gamma(q)/Gamma(p+q).

   return TMath::Exp(TMath::LnGamma(p)+TMath::LnGamma(q)-TMath::LnGamma(p+q));
}

  Double_t OpeTool::BetaCf(Double_t x, Double_t a, Double_t b)
    {
      // Continued fraction evaluation by modified Lentz's method
      // used in calculation of incomplete Beta function.
      
      Int_t itmax = 500;
      Double_t eps = 3.e-14;
      Double_t fpmin = 1.e-30;
      
      Int_t m, m2;
      Double_t aa, c, d, del, qab, qam, qap;
      Double_t h;
      qab = a+b;
      qap = a+1.0;
      qam = a-1.0;
      c = 1.0;
      d = 1.0 - qab*x/qap;
      if (TMath::Abs(d)<fpmin) d=fpmin;
      d=1.0/d;
      h=d;
      for (m=1; m<=itmax; m++) {
	m2=m*2;
	aa = m*(b-m)*x/((qam+ m2)*(a+m2));
	d = 1.0 +aa*d;
	if(TMath::Abs(d)<fpmin) d = fpmin;
	c = 1 +aa/c;
	if (TMath::Abs(c)<fpmin) c = fpmin;
	d=1.0/d;
	h*=d*c;
	aa = -(a+m)*(qab +m)*x/((a+m2)*(qap+m2));
	d=1.0+aa*d;
	if(TMath::Abs(d)<fpmin) d = fpmin;
	c = 1.0 +aa/c;
	if (TMath::Abs(c)<fpmin) c = fpmin;
	d=1.0/d;
	del = d*c;
	h*=del;
	if (TMath::Abs(del-1)<=eps) break;
      }
      if (m>itmax) {
	throw Sct::LogicError("OpeTool::BetaCf: a or b too big, or itmax too small", __FILE__, __LINE__);
      }
      return h;
    }
}
