#include "CalibrationControllerImpl.h"
#include "ScanLibraryImpl.h"
#include "TestLibraryImpl.h"
#include "SequenceLibraryImpl.h"
#include "SequenceRequestWorker.h"
#include "RunController.h"
#include "SctApiAccessException.h"
#include "ipc/SequenceRequest.h"
#include "scripts/DefaultSequence.h"
#include "scripts/DefaultTest.h"
#include "Sequence.h"
#include "Test.h"
#include "SctApiCall.h"
#include "ConfigUpdater/ConfigUpdaterManager.h"

#include "Sct/SctNames.h"
#include "Sct/LogicErrors.h"
#include "Sct/IoExceptions.h"
#include "Sct/ISUtilities.h"
#include "Sct/CorbaExceptionWrapper.h"
#include "Sct/StdExceptionWrapper.h"
#include "Sct/IS/IOManagerIS.h"
#include "Sct/IS/IONameIS.h"
#include "Sct/OmniMarshalling.h"
#include "SctData/TestResult.h"
#include "Sct/UniqueID.h"
#include <ipc/server.h>
#include <is/isinfotmpl.h>
//#include <ilu/iluxport.h>
#include <boost/utility.hpp>
#include <unistd.h>
#include <iostream>
#include "Sct/BugReport.h"
#include "Sct/CorbaExceptionWrapper.h"

#ifndef __Sct_SctApi_H_
#include "Sct_SctApi/SctApi.hh"
#endif


using namespace std;
using namespace Sct;
using namespace SctData;

namespace SctCalibrationController {

using namespace Scripts;

CalibrationControllerImpl::CalibrationControllerImpl(RunControl& rc) : rc(rc), runNumber(0), 
    nextScanNumber(0), abortNow(false), abortRightNow(false) {
    status.currentScanIndex = 0;
    publish();
    updateStatus();
}

CalibrationControllerImpl::~CalibrationControllerImpl() {
    withdraw();

    //Remove status from IS:
    string name = SctNames::getControlDataName();
    name += ".Status";
    ISInfoDictionary& is = SctNames::getISDictionary();
    is.remove(name.c_str());
    inst->_destroy();
}

Sct_SctApi::SctApiIPC_ptr CalibrationControllerImpl::getApi() {
  if (CORBA::is_nil(m_api)){
    // get new handle...
    m_api=SctNames::getPartition().lookup<Sct_SctApi::SctApiIPC>(Sct_SctApi::SctApiIPC::instanceName);
    if (CORBA::is_nil(m_api)) {
      SctNames::Mrs() << "CC" << MRS_TEXT("Calibration Controller failed to find SctApi") << MRS_ERROR << ENDM;
    } else {
      SctNames::Mrs() << "CC" << MRS_TEXT("Calibration Controller refreshed SctApi handle") << MRS_DIAGNOSTIC << ENDM;
    }
  }
  Sct_SctApi::SctApiIPC_ptr ret = Sct_SctApi::SctApiIPC::_duplicate(m_api);
  if (CORBA::is_nil(ret)) {
      SctNames::Mrs() << "CC" << MRS_TEXT("Calibration Controller failed to duplicate handle") << MRS_ERROR << ENDM;    
  };
  return ret;
}

void CalibrationControllerImpl::setApi(Sct_SctApi::SctApiIPC_ptr api) {
    this->m_api = Sct_SctApi::SctApiIPC::_duplicate(api);
}

void CalibrationControllerImpl::reset() {
    SctNames::Mrs() << "CC_RESET" << MRS_TEXT("Calibration controller was reset") << MRS_DIAGNOSTIC << ENDM;
    status.status = status.LOADED;
    updateStatus();
    m_api = Sct_SctApi::SctApiIPC::_nil();
}

void CalibrationControllerImpl::sctApiError(SctApiAccessException& e) {
  rc.sctApiError(/*rc.SCTAPI_ACCESS_FAILURE, */e.what());
  //status.status = status.ERROR;
  updateStatus();
}

void CalibrationControllerImpl::takeControl(unsigned int runNumber) {
    if (status.status != status.LOADED || CORBA::is_nil(m_api) )
        return;
    this->runNumber = runNumber;
    nextScanNumber = 0;
    status.status = status.INCONTROL;
    
    SctNames::Mrs() << "CC_START" << MRS_TEXT("Calibration Controller taking control") << MRS_INFORMATION << ENDM;
    updateStatus();
}

void CalibrationControllerImpl::giveupControl() {
    if (status.status == status.INCONTROL) {
	SctNames::Mrs() << "CC_END" << MRS_TEXT("Calibration Controller giving up control") << MRS_INFORMATION << ENDM;

        status.status = status.LOADED;
        updateStatus();
    }
}

bool CalibrationControllerImpl::isInControl() {
    return status.status == status.INCONTROL;
}

bool CalibrationControllerImpl::isBusy() {
    return status.status == status.BUSY;
}

void CalibrationControllerImpl::setBusy(bool busy) {
    if (busy && status.status == status.INCONTROL)
        status.status = status.BUSY;
    else if (!busy && status.status == status.BUSY)
        status.status = status.INCONTROL;
    updateStatus();
}

unsigned int CalibrationControllerImpl::getRunNumber() const {
    return runNumber;
}

unsigned int CalibrationControllerImpl::getNextScanNumber() const {
    return nextScanNumber;
}

void CalibrationControllerImpl::setNextScanNumber(unsigned int scanNumber) {
    nextScanNumber = scanNumber;
    Sct_SctApi::SctApiIPC_var api = getApi();
    APICALL(api, setScanNumber(scanNumber), "setScanNumber failed in setNextScanNumber")
}

void CalibrationControllerImpl::abort() {
  /// If the CC is not busy there is nothing to abort, so ignore:
    if (status.status != status.BUSY) return;

    abortNow = true;
    //Notify MRS
    SctNames::Mrs() << "CC_ABORT" << MRS_TEXT("User requested abort Sequence") << MRS_INFORMATION << ENDM;

    //See if SctApi is responding, if not, then abortRightNow!
    //Use pid as an example...
    try{
      Sct_SctApi::SctApiIPC_var api = getApi();
      ipc::servant::info_var info = api->information();
    }catch(CORBA::Exception& e){
      abortRightNow = true;
    }
}

void CalibrationControllerImpl::executeSequence(Sct_CalibrationController::SequenceRequest_ptr sr) {
    SctNames::Mrs() << "CC_SEQ_START" << MRS_TEXT("Calibration Controller starting sequence") << MRS_INFORMATION << ENDM;

    string name;
    Sct_SctApi::SctApiIPC_var api = getApi();

    Sct::BugReport("BUG01", "CCImpl.cpp", "Just before startSequence try block");
    try {
      Sct::BugReport("BUG01", "CCImpl.cpp", "Just before xx1 startSequence try block");
      if (CORBA::is_nil(sr)) {
        Sct::BugReport("BUG01", "CCImpl.cpp", "Oh dear, sr was nil!");
        throw Sct::IllegalStateError("Null Sequence pointer ", __FILE__, __LINE__);
      };
      Sct::BugReport("BUG01", "CCImpl.cpp", "Just before xx2 startSequence try block");
//First up, synchronise our scanNumber with SctApi's
      APIRETCALL(api, nextScanNumber, getScanNumber(), "Couldn't synchronise scan numbers")
      //The call startSequence
      Sct::BugReport("BUG01", "CCImpl.cpp", "Just before xx3 startSequence try block");
      CORBA::String_var corbaname = sr->startSequence(api);
      Sct::BugReport("BUG01", "CCImpl.cpp", "Just before xx4 startSequence try block");
      name = corbaname;
      Sct::BugReport("BUG01", "CCImpl.cpp", "Just before xx5 the startSequence we are trying to understand.");
      std::cout << "CC: name of sequence is : " << name << std::endl;
    } catch (SctApiAccessException& e) {
      Sct::BugReport("BUG01", "CCImpl.cpp", "In first catch");
      sctApiError(e);
      throw;	//Drop through
    } catch (CorbaExceptionWrapper& e) {
      Sct::BugReport("BUG01", "CCImpl.cpp", "In second catch");
      e.sendToMrs();
      return;
    }
    Sct::BugReport("BUG01", "CCImpl.cpp", "posn 2");

    shared_ptr<Sequence> s;
    Sct::BugReport("BUG01", "CCImpl.cpp", "posn 3");
    try {
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 4");
        sctConf::ModuleList_var mlist;
	std::cout << "CC Getting module List" << std::endl;
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 5");
        APIRETCALL(api, mlist, getModuleList(), "Couldn't get module list")
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 6");
	std::cout << "CC Got module List" << std::endl;
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 7");

        list<string> list = copyCorbaToList<sctConf::ModuleList, string>(&(mlist.in()));
        s = Sequence::create(name, runNumber, nextScanNumber, list);

        status.currentSequence = s->getUniqueID();
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 8");
        updateStatus();
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 9");

        Sct_CalibrationController::TestRequest_var tr = Sct_CalibrationController::TestRequest::_nil();
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 10");
        unsigned long index = 0;

        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 11");
	do {
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 12");
	    tr = sr->getNextTest(index);
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 12.5");
	    if (!CORBA::is_nil(tr)) {
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 13");
		unsigned firstScanInTest = getNextScanNumber();
		unsigned runNumber = getRunNumber();
		s->addTest(executeTest(tr, s));
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 14");

		//Check we didn't fail somewhere
		if (s->getTest(index)->getData().status != TestData::COMPLETED) {
		    SctNames::Mrs() << "CC_SEQ_END" << MRS_TEXT("Calibration Controller aborting sequence") << MRS_INFORMATION << ENDM;
		    s->setStatus(SequenceData::ABORTED);
		    return;	//All done!
		}
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 15");

		if (tr->canFeedback()) {
		    std::list<std::string> moduleList = s->getModuleList();
		    applyChanges(runNumber, firstScanInTest, moduleList);
		}
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 16");
	    } else {
              Sct::BugReport("BUG01", "CCImpl.cpp","posn 16.5 -- tr was null!");
            };
	    ++index;
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 17");
	} while (!CORBA::is_nil(tr));
	
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 18");
        // Save any changes that have been made (into the ROD)
        APICALL(api, setABCDModules(Sct_SctApi::CALIBRATION_CONFIG), "setABCDModules failed after update")
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 19");

	SctNames::Mrs() << "CC_SEQ_END" << MRS_TEXT("Calibration Controller ending sequence") << MRS_INFORMATION << ENDM;
	sr->endSequence();
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 20");
	s->setStatus(SequenceData::COMPLETED);
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 21");

    } catch (SctApiAccessException& e) {
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 22");
	sctApiError(e);
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 23");
	throw;	//Drop through
    } catch (Throwable& e) {
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 24");
	e.sendToMrs(MRS_INFORMATION);
	if (s) s->setStatus(SequenceData::ABORTED);
    }
        Sct::BugReport("BUG01", "CCImpl.cpp", "posn 25");
}

void CalibrationControllerImpl::applyChanges(const unsigned long runNumber, const unsigned long scanNumber, const std::list<std::string>& modulelist ) {
    Sct_SctApi::SctApiIPC_var api = getApi();

    cout << "Update option = " << status.updateOption << endl;

    bool printed=false;
    while (status.updateOption=="WAIT") {
        if (!printed)
            cerr << "CalibrationController awaiting instructions as to whether to apply changes" << endl;
        printed=true;
    }

    //status.updateOption can be changed by other IPC method
    if (status.updateOption!="UPDATE")
        return;

    cout << "updating " << modulelist.size() << "" << " modules" << endl;

    std::ostringstream regexp;
    regexp << ".*TestResult\\." << runNumber << "\\." << scanNumber << "\\..*";
    cout << "Cal Controller looking for " << regexp.str() << endl;

    unsigned nprinted=0;

    // make another list which we can remove things from:
    std::list<std::string> mylist=modulelist;

    while (!mylist.empty() && status.updateOption=="UPDATE") {
        ISInfoIterator ii(SctNames::getPartition(),
                          SctNames::getTestDataName().c_str(),
                          regexp.str().c_str() );

        while (ii() && status.updateOption=="UPDATE") {
            if (nprinted<10)
                cout << "    checking " << ii.name() << endl;


            // make an IONameIS which allows us to interpret the IS name
            Sct::IS::IONameIS isname(ii.name());
            std::string moduleName=isname.getUniqueID().getModule();

            // look for the module name in our list
            std::list<string>::iterator it = std::find( mylist.begin(), mylist.end(), moduleName);
            // if its there, then update that module, and remove it from our list.
            if ( it != mylist.end() ) {
                shared_ptr<const Serializable> ob ( Sct::IS::IOManagerIS::instance().read(ii.name()) );
                shared_ptr<const SctData::TestResult> testResult = boost::dynamic_pointer_cast<const SctData::TestResult>(ob);
                if (!testResult)
                    throw Sct::InvalidArgumentError("Fitter::applyChangesCallback not a SctData::TestResult.", __FILE__, __LINE__);

                cout << "about to update " << moduleName << endl;
		ConfigUpdaterManager::instance().update(*testResult, api);
                it=mylist.erase(it);
            }
        } // end of loop over currently available TestResult

        sleep (1);

        if (nprinted<10) {
            cout << "Cal Controller waiting for TestResults (have "
            << mylist.size() << " still to go)" << endl;
        }
        nprinted ++;
    }
}

void CalibrationControllerImpl::updateWith(const char * testResultInIs, bool force){
  setBusy(true);
  try{
    Sct_SctApi::SctApiIPC_var api = getApi();
    shared_ptr<const TestResult> tr = dynamic_pointer_cast<const TestResult> ( Sct::IS::IOManagerIS::instance().read(testResultInIs) );
    if (!tr.get()) throw Sct::IoException(string("Couldnt retrieve ")+testResultInIs , __FILE__, __LINE__);
    ConfigUpdaterManager::instance().update(*tr, api, force);
  }catch(Sct::Throwable& e){
    e.sendToMrs();
  }catch(std::exception& e){
    Sct::StdExceptionWrapper(e, __FILE__, __LINE__).sendToMrs();
  }
  setBusy(false);
}

void CalibrationControllerImpl::setUpdateOption(Sct_CalibrationController::CalibrationController::UpdateOption opt) {
    cout << "CalibrationContrllerImpl: resetting update option to " << opt;
    switch (opt) {
    case Sct_CalibrationController::CalibrationController::update : {
        status.updateOption="UPDATE";
        SctNames::Mrs() << "CC_UPDATE_CHANGE"
			<< MRS_TEXT("CalibrationControllerImpl setting updateOption=UPDATE") << MRS_DIAGNOSTIC << ENDM;
        break;
    }
    case Sct_CalibrationController::CalibrationController::noupdate : {
	status.updateOption="NOUPDATE";
	SctNames::Mrs() << "CC_UPDATE_CHANGE"
			<< MRS_TEXT("CalibrationControllerImpl setting updateOption=NOUPDATE")  << MRS_DIAGNOSTIC << ENDM;
	break;
    }
    case Sct_CalibrationController::CalibrationController::wait : {
	status.updateOption="WAIT";
	SctNames::Mrs() << "CC_UPDATE_CHANGE"
			<< MRS_TEXT("CalibrationControllerImpl setting updateOption=WAIT")  << MRS_DIAGNOSTIC << ENDM;
	break;
    }
    default : {
	IllegalStateError e("CalibrationControllerImpl updateOption UNKNOWN", __FILE__, __LINE__);
	e.sendToMrs(MRS_DIAGNOSTIC);
	break;
    }
    }
    updateStatus();
}

auto_ptr<Test> CalibrationControllerImpl::executeTest(Sct_CalibrationController::TestRequest_ptr tr, shared_ptr<const Sequence> s) {
    BugReport("BUG01", "CCImpl.cpp","Starting executeTest");
 
    BugReport("BUG01", "CCImpl.cpp","In executeTest 1");
    TestData data;
     BugReport("BUG01", "CCImpl.cpp","In executeTest 2");
    BugReport("BUG01", "CCImpl.cpp","In executeTest 3");

    if (CORBA::is_nil(tr)) {
      BugReport("BUG01", "CCImpl.cpp","Ending executeTest early 0");
      throw IllegalStateError("Test request pointer was nill", __FILE__, __LINE__);
    };

    SctNames::Mrs() << "CC_TEST_START" << MRS_TEXT("Calibration Controller starting test") << MRS_INFORMATION << ENDM;
    BugReport("BUG01", "CCImpl.cpp","In executeTest 4");
    Sct_SctApi::SctApiIPC_var api = getApi();
    BugReport("BUG01", "CCImpl.cpp","In executeTest 5");
    APICALL(api, setABCDModules(Sct_SctApi::CALIBRATION_CONFIG), "setABCDModules failed")
    BugReport("BUG01", "CCImpl.cpp","In executeTest 6");

    data.runNumber = runNumber;
    BugReport("BUG01", "CCImpl.cpp","In executeTest 7");
    data.startScanNumber = nextScanNumber;
    BugReport("BUG01", "CCImpl.cpp","In executeTest 8");

    {
      Sct_CalibrationController::TestPoints_var testPointsOut;
      {
        BugReport("BUG01", "CCImpl.cpp","In executeTest 9");
        //TestRequestAdapter tra(tr);
        BugReport("BUG01", "CCImpl.cpp","In executeTest 10");
        CORBA::String_var testNameOut;
        BugReport("BUG01", "CCImpl.cpp","In executeTest 10.1");
        CORBA::UShort testVariableOut;
        BugReport("BUG01", "CCImpl.cpp","In executeTest 10.2");
        CORBA::ULong nScansOut;
        BugReport("BUG01", "CCImpl.cpp","In executeTest 10.3");
        tr->startTest(api, testNameOut, testVariableOut, nScansOut, testPointsOut);
        if (CORBA::is_nil(tr)) {
          BugReport("BUG01", "CCImpl.cpp","Ending executeTest early 000 .. tr was accidentally made null!");
          throw IllegalStateError("Test request pointer was accidentally made nill at ", __FILE__, __LINE__);
        };
        BugReport("BUG01", "CCImpl.cpp","In executeTest 10.4");
        data.testName = testNameOut;
        BugReport("BUG01", "CCImpl.cpp","In executeTest 10.5");
        data.testVariable = testVariableOut;
        BugReport("BUG01", "CCImpl.cpp","In executeTest 10.6");
        data.nScans = nScansOut;
        BugReport("BUG01", "CCImpl.cpp","In executeTest 11.7");
      };
      
      BugReport("BUG01", "CCImpl.cpp","In executeTest 12");
      data.testPoints = new double[testPointsOut->length()];
      BugReport("BUG01", "CCImpl.cpp","In executeTest 13");
      data.testPoints_size = testPointsOut->length();
      BugReport("BUG01", "CCImpl.cpp","In executeTest 14");
      for (unsigned int i=0; i<testPointsOut->length(); ++i) {
        BugReport("BUG01", "CCImpl.cpp","In executeTest 15");
        data.testPoints[i] = testPointsOut[i];
        BugReport("BUG01", "CCImpl.cpp","In executeTest 16");
      }
    };

    BugReport("BUG01", "CCImpl.cpp","In executeTest 17");
    auto_ptr<Test> test (new Test(data, s->getModuleList()));
    BugReport("BUG01", "CCImpl.cpp","In executeTest 18");

    status.currentTest = test->getUniqueID();
    BugReport("BUG01", "CCImpl.cpp","In executeTest 19");
    updateStatus();
    BugReport("BUG01", "CCImpl.cpp","In executeTest 20");

    try {
    BugReport("BUG01", "CCImpl.cpp","In executeTest 21");
    	unsigned int i=0;
	bool errorFlag=false;
	for (; i<data.nScans; ++i) {
    BugReport("BUG01", "CCImpl.cpp","In executeTest 22");
	    Sct_CalibrationController::ScanRequest_var scan = tr->getNextScan(i);
    BugReport("BUG01", "CCImpl.cpp","In executeTest 22.5");
	    if (CORBA::is_nil(scan)) {
    BugReport("BUG01", "CCImpl.cpp","In executeTest 23");
              throw IllegalStateError("getNextScan returned zero!", __FILE__, __LINE__);
            };
    BugReport("BUG01", "CCImpl.cpp","In executeTest 23.5");
	    if (CORBA::is_nil(scan->getScanCorba())) {
    BugReport("BUG01", "CCImpl.cpp","In executeTest 24");
              throw IllegalStateError("ScanRequest.getScan returned zero!", __FILE__, __LINE__);
            };
    BugReport("BUG01", "CCImpl.cpp","In executeTest 25");
	    test->addScan(scan->getScanCorba());
    BugReport("BUG01", "CCImpl.cpp","In executeTest 26");
	    if (abortNow) {
    BugReport("BUG01", "CCImpl.cpp","In executeTest 27");
		abortNow = false;
		abortRightNow = false;
		break;
	    }
    BugReport("BUG01", "CCImpl.cpp","In executeTest 28");
	    try{
    BugReport("BUG01", "CCImpl.cpp","In executeTest 29");
	      executeScan(scan, i);
    BugReport("BUG01", "CCImpl.cpp","In executeTest 30");
	    }catch(CORBA::Exception& e){
    BugReport("BUG01", "CCImpl.cpp","In executeTest 31");
	      throw CorbaExceptionWrapper(e,"Exception in executeScan", __FILE__, __LINE__);
	    }
    BugReport("BUG01", "CCImpl.cpp","In executeTest 32");
	}
    BugReport("BUG01", "CCImpl.cpp","In executeTest 33");
	if (i == data.nScans && !errorFlag) {
    BugReport("BUG01", "CCImpl.cpp","In executeTest 34");
	    SctNames::Mrs() << "CC_TEST_END" << MRS_TEXT("Calibration Controller ending test") << MRS_INFORMATION << ENDM;
	    test->setStatus(TestData::COMPLETED);
    BugReport("BUG01", "CCImpl.cpp","In executeTest 35");
	} else {
    BugReport("BUG01", "CCImpl.cpp","In executeTest 36");
	    SctNames::Mrs() << "CC_TEST_ABORT" << MRS_TEXT("Calibration Controller aborting test") << MRS_INFORMATION << ENDM;
	    test->setStatus(TestData::ABORTED);
    BugReport("BUG01", "CCImpl.cpp","In executeTest 37");

	    const TestData& testData = test->getData();
	    setNextScanNumber(testData.startScanNumber + testData.nScans);
	}
    BugReport("BUG01", "CCImpl.cpp","In executeTest 38");

    } catch (SctApiAccessException& e) {
    BugReport("BUG01", "CCImpl.cpp","In executeTest 39");
	sctApiError(e);
        BugReport("BUG01", "CCImpl.cpp","Ending executeTest early 2");
	throw;	//Drop through
    } catch (CorbaExceptionWrapper& e) {
	SctNames::Mrs() << "CC_TEST_ABORT" << MRS_TEXT("Calibration Controller aborting test") << MRS_INFORMATION << ENDM;
	test->setStatus(TestData::ABORTED);
    BugReport("BUG01", "CCImpl.cpp","In executeTest 40");
	e.sendToMrs(MRS_INFORMATION);

	//Try an increment the scan number to the end of the test
	const TestData& testData = test->getData();
    BugReport("BUG01", "CCImpl.cpp","In executeTest 41");
	setNextScanNumber(testData.startScanNumber + testData.nScans);
    }
    BugReport("BUG01", "CCImpl.cpp","In executeTest 42");

    //End the test
    try {
    BugReport("BUG01", "CCImpl.cpp","In executeTest 43");
    	tr->endTest();
    BugReport("BUG01", "CCImpl.cpp","In executeTest 44");
    } catch (CorbaExceptionWrapper& e) {
    BugReport("BUG01", "CCImpl.cpp","In executeTest 45");
	e.sendToMrs(MRS_DIAGNOSTIC);
    BugReport("BUG01", "CCImpl.cpp","In executeTest 46");
    }
    BugReport("BUG01", "CCImpl.cpp","In executeTest 47");

    //Reset the module configurations to the original
    try {
    BugReport("BUG01", "CCImpl.cpp","In executeTest 48");
        APICALL(api, getABCDModules(Sct_SctApi::CALIBRATION_CONFIG), "getABCDModules failed")
    BugReport("BUG01", "CCImpl.cpp","In executeTest 49");
        APICALL(api, sendABCDModules(Sct_SctApi::PHYSICS_CONFIG), "sendABCDModules failed")
    BugReport("BUG01", "CCImpl.cpp","In executeTest 50");
    } catch (SctApiAccessException& e) {
    BugReport("BUG01", "CCImpl.cpp","In executeTest 51");
        sctApiError(e);
    BugReport("BUG01", "CCImpl.cpp","In executeTest 52");
        e.sendToMrs();
    BugReport("BUG01", "CCImpl.cpp","In executeTest 53");
    }
    BugReport("BUG01", "CCImpl.cpp","In executeTest 54");

    BugReport("BUG01", "CCImpl.cpp","Ending executeTest normally");
 return test;
}

void CalibrationControllerImpl::executeScan(Sct_CalibrationController::ScanRequest_ptr scan, unsigned int index) {
    Sct_SctApi::SctApiIPC_var api = getApi();
    ++nextScanNumber;
    SctNames::Mrs() << "CC_SCAN_START" << MRS_TEXT("Calibration Controller starting scan") << MRS_INFORMATION << ENDM;

    status.currentScanIndex = index;
    updateStatus();

    SctNames::Mrs() << "CC_SCAN_PART1" << MRS_TEXT("Calibration Controller determining whether scan is raw or cooked ...") << MRS_INFORMATION << ENDM;
    if ( scan->isRawCorba() ){
      SctNames::Mrs() << "CC_SCAN_PART2" << MRS_TEXT("Calibration Controller thinks scan is raw.  Now doingRawScan ...") << MRS_INFORMATION << ENDM;
      APICALL(api, doRawScan(scan->getScanCorba(), scan->delayCorba(), scan->widthCorba(), scan->configureModulesCorba(), scan->clockByTwoCorba()  ), "doRawScan failed");
    }else{
      SctNames::Mrs() << "CC_SCAN_PART2" << MRS_TEXT("Calibration Controller thinks scan is cooked.  Now doingCookedScan ...") << MRS_INFORMATION << ENDM;
      APICALL(api, doScan(scan->getScanCorba() ), "doScan failed");
    }
    SctNames::Mrs() << "CC_SCAN_PART3" << MRS_TEXT(" ... finished whatever kind of scan we started.") << MRS_INFORMATION << ENDM;

    // Perhaps the next few lines could be replaced by an APICALL to awaitScan ??   CGL: 2004/01/24
    // No, awaitScan has no timeout, or other get out (except at CORBA level)
    // CalibrationController may want to give more feedback here about what is going on. BJG: 2004/01/25
    // Also would be place to put APICALL to abortScan, once implemented properly? 

    ISInfoInt i;
    i.setValue(-1);
    ISInfo::Status isStatus = ISInfo::Success;
    ISInfoDictionary& is = SctNames::getISDictionary();
    int nInitialSleeps = 0;

    while ((i.getValue() == -1 && nInitialSleeps < 20) || (i.getValue() == 1 && isStatus == ISInfo::Success)) {
        isStatus = is.findValue("SCTAPIServer.scanning", i);
        usleep(100);
	++nInitialSleeps;
	if (abortRightNow) {
	    abortRightNow = false;
	    abortNow = false;
	    IoException erm("Abort called and SctApi not responding, doScan failed", __FILE__, __LINE__);
	    throw erm;
	}
    }

    SctNames::Mrs() << "CC_SCAN_END" << MRS_TEXT("Calibration Controller ending scan") << MRS_INFORMATION << ENDM;
}

  CalibrationControllerImpl & CalibrationControllerImpl::initialize(RunControl& rc) {
    inst = new CalibrationControllerImpl(rc);
    return *inst;
}

Sct_CalibrationController::CalibrationController_ptr CalibrationControllerImpl::instance() {
    if (!inst)
	throw InvariantViolatedError("Need to call CalibrationControllerImpl::initialize first!", __FILE__, __LINE__);
    return inst->_this();
}

Sct_CalibrationController::ScanLibrary_ptr CalibrationControllerImpl::getScanLibrary() {
    return ScanLibraryImpl::instance();
}

Sct_CalibrationController::TestLibrary_ptr CalibrationControllerImpl::getTestLibrary() {
    return TestLibraryImpl::instance();
}

Sct_CalibrationController::SequenceLibrary_ptr CalibrationControllerImpl::getSequenceLibrary() {
    cout << "Get Sequence Library" << endl;
    return SequenceLibraryImpl::instance();
}

void CalibrationControllerImpl::doScan(Sct_CalibrationController::ScanRequest_ptr scan) {
    doTest(DefaultTest::instance(scan));
}

void CalibrationControllerImpl::doTest (Sct_CalibrationController::TestRequest_ptr t) {
    doSequence(DefaultSequence::instance(t));
}

void CalibrationControllerImpl::doSequence (Sct_CalibrationController::SequenceRequest_ptr r) {
    if (!isInControl() || isBusy()) {
        cerr << "Can't execute SequenceRequest - bad status: " << status.status << endl;
        return;
    }

    cout << "do Sequence" << endl;

    status.status = status.BUSY;  // set back to INCONTROL by the SequenceRequestWorker
    updateStatus();


    m_sequence_request_thread_group.create_thread(SequenceRequestWorker(*this, r));
    Sct::BugReport("WARN01","CC","the second argument to this func should probably inherit from the self-destructing portable adapter mixin class ...");
#warning the second argument to this func should probably inherit from the self-destructing portable adapter mixin class ...
}

Sct_SctApi::Scan_ptr CalibrationControllerImpl::getScan(unsigned long runNumber, unsigned long scanNumber) {
  cerr << "CalibrationControllerImpl::getScan - Not implemented" << endl;
  return Sct_SctApi::Scan::_nil();
}

void CalibrationControllerImpl::updateStatus() {
  Sct::BugReport("BUG01","CCImpl.cpp","In updateStatus 1/4");
  string name = SctNames::getControlDataName();
  Sct::BugReport("BUG01","CCImpl.cpp","In updateStatus 2/4");
  name += ".Status";
  Sct::BugReport("BUG01","CCImpl.cpp","In updateStatus 3/4");
  ISUtilities::addOrUpdateOrMessage(name, status, __FILE__, __LINE__, MRS_DIAGNOSTIC);
  Sct::BugReport("BUG01","CCImpl.cpp","In updateStatus 4/4");
}
  
CalibrationControllerImpl * CalibrationControllerImpl::inst = 0;

}
