#ifndef FITTER_H
#define FITTER_H

#include <is/isinfo.h>
#include <ipc/object.h>
#include <ipc/server.h>

#include <string>
#include <memory>

#include <boost/thread.hpp>

#include "FitterInterface.hh"
#include "FitterArguments.h"
#include "Sct/Exception.h"
#include "Sct/ConfigurationException.h"
#include "Sct/LogicErrors.h"
#include "Sct/IoExceptions.h"

class TH1;
class TF1;

using namespace Sct;

/**
 * Containing code specific to the FittingService. Uses objects from SctData and Sct.
 */
namespace SctFitter {
    class FitStrategy;
    class FitterWorkerGroup;

/**
 * Class definition of the SCT fitting service.
 * This service subscribes to an input IS server, receiving new RawScanResult objects.
 * It then fits a function (dependant on the ScanType) to chips, strips, etc.
 * and outputs a FitScanResult object to the output IS server.<BR>
 * Some status information about the Fitter is available via IPC request.
 * @author Alan Barr
 * @date 16 December 2002
 */
class Fitter:
  public IPCNamedObject<POA_FitterI::FitterInterface,
                        ipc::multi_thread>,
  private boost::noncopyable {
private:
  /// The following method is a new requirement of IPCNamedObjects under the omni-orb
  void shutdown();
  /// singleton
  Fitter(FitterArguments args) throw(ConfigurationException);
  /// private destructor
  ~Fitter() throw();
  static Fitter* fitter;
    
public:
    /**
      Gets a reference to the instance of the Fitter.
      If it hasn't been initialized, it calls initialize() first
    */
    static Fitter& instance();
    
    /**
      Initialize the Fitter instance
      @param fitStrategyName the name of the FitStrategy to use.  Default is RootFitStrategy
      @throws ConfigurationException if the fitStrategyName is not a recognized FitStrategy
      */
    static Fitter& initialize(FitterArguments args) throw(ConfigurationException);
    
    /**
     * The command to start the workers.
     */
    void go() throw(IsException);
  private:
    // INTERNAL METHODS:
    /**
     * Get the status information
     */
    const char* internal_status() const throw();
    /**
     * Get the name of the last scan which was fitted;
     */
    const char* internal_lastScan() const throw();
    /// get the strategy. Throws an error if none is defined.
  public:
    FitStrategy& internal_getFitStrategy() const throw(LogicError) ;
  private:
    void internal_setFitOptions(const string & opt) throw(LogicError) ;
    /**
    Set the stragegy from a name the the FitStrategy map.
    @throw LogicError if that name is not in the map.
    */
    void internal_setFitStrategy(const string & name) throw(LogicError) ;
    /**
     *  returns the number of fits done so far;
     */
    long internal_nFitsDone() const throw();
    /**
     *  returns the number of fit errors so far;
     */
    long internal_nFitErrors() const throw();
    /**
      Return true if we're using analytic fit algorithms
      */
    bool internal_isUsingAnalyticAlgorithm() const;
    /**
     * The infoReceiver callback function must take a static function as an argument.
     * This function simply calls doFits on the parameter or `rock', which is set 
     * to be `this' (Fitter). <BR>
     * At the moment it catches all errors since I think it should still
     * be in a stable state after most logic errors I can think of. AJB
     */
    static void doFits(ISCallbackInfo * isc) throw(IsException, LogicError);
    
  public:
    // CORBA METHODS:
    /**
     * As for getStatus, but available over IPC. Makes a new char[] which ipc then deletes.
     * @param the status of the IPC request.
     */
    virtual char* status() throw();
    /**
     * Get the name of the last scan which was fitted over ipc;
     */
    char* lastScan() throw();
    /**
     * Finds the FitStrategy options. Makes a new char[] which ipc then deletes.
     */
    virtual char* getFitOptions() throw() ;
    /**
     * Finds the FitStrategy name. Makes a new char[] which ipc then deletes.
     */
    virtual char* getFitStrategy() throw() ;
    /**
     * set the fit option over IPC.
     * (asynchronous).
     * Passes string on to FitStrategy;
     * @param the status of the IPC request.
     * @param the string to set as the options e.g. "NQR" for ROOT.
     */
    /// as for setFitOptions, but over IPC.
    virtual void setFitOptions(const char* opt) throw() ;
    /**
     * set the fit option over IPC.
     * (asynchronous).
     * @param the status of the IPC request : failure if name not found in map.
     * @param tries to find the stragegy with this name 
     */
    virtual void setFitStrategy(const char* strategy) throw() ;
    // Fit all RawScanResults currently in IS.
    void fitAll() throw();
    /// Fit a named RawScanResult from IS.
    void fit(const char* name) throw();
    /** number of busy workers */
    long busy();
    /** number of jobs in the queue */
    long queueLength();
    /**
     *  returns the number of fits done so far over ipc;
     */
    long nFitsDone() throw();
    /**
    *  returns the number of fit errors so far over ipc;
    */
    long nFitErrors() throw();
    /**
      Determine whether to use the analytic fit algorithm.  Default is false
      */
    void useAnalyticAlgorithm (CORBA::Boolean use);
    /**
      Return true if we're using analytic fit algorithms
      */
    CORBA::Boolean isUsingAnalyticAlgorithm();

    
    /// OTHER METHODS .........
  public:
    /**
       Increments the number of fit errors by 1.
    */
    void incrementFitErrors() throw();
    
    /**
       Increments the number of fits done by 1.
    */
    void incrementFitsDone() throw();
    
    /**
      Called by the workers to add to the total amount of scan time
      */
    void scanDone(double time) throw();
    
    /**
      Called to add to IO time.  Assumes that main thread does IO and is running
      concurrently with workers.
      */
    void addIOTime(double time) throw();
    
    /**
      Algorithm may be complex!  Check actual code
      */
    double getAverageTimePerScan() const throw();
    
    /**
       Server to write fits to...
    */
    std::string getOutputIsServer() const;
    
    /**
    Get the Fitter IPC server
    */
    static IPCServer& getFitterServer() throw();
    
private:      

    /*
     * counters:      
     */
    long m_nFitsDone;
    long m_nFitErrors;
    long m_nScans;
    double m_scanTimeTaken;
    double m_ioTimeTaken;
    ///Used to lock access to all the counters
    mutable boost::recursive_mutex counterMutex;
    
    /*
     * name of the last scan attempted.
     */
    string m_scan;
    /**
     * pointer to char containing the status: cached so that we don't have to make lots of char*'s.
     */
    mutable string m_status;
    /**
     * Pointer to the fit strategy.
     */
    FitStrategy* m_fitStrategy;
    /// Group of workers which do the actual fitting.
    FitterWorkerGroup* workergroup;    
    /**
     * An InfoReveiver for RawRODData objects
     */
    std::auto_ptr<ISInfoReceiver> m_infoReceiver;

    /// update record of last scan attempted.
    void setLastScanName(const std::string& name){
	m_scan = name;
    }

    void updateSubscriptions();

    /// FitterWorker is allowed to update stuff.
    friend class FitterWorkerGroup;
    FitterArguments m_args;
}
; // end of definition of class Fitter;
}// end of namespace SctFitter
#endif // #ifndef FITTER_H
